@@ -1,10 +1,10 @@
-/*  Plzip - A parallel compressor compatible with lzip
+/*  Plzip - Parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2009-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -45,17 +45,17 @@
 //
 int readblock( const int fd, uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = read( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
+    const int n = read( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
     else if( n == 0 ) break;				// EOF
-    else if( errno != EINTR && errno != EAGAIN ) break;
+    else if( errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
@@ -64,77 +64,86 @@ int readblock( const int fd, uint8_t * const buf, const int size )
 //
 int writeblock( const int fd, const uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = write( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
-    else if( n < 0 && errno != EINTR && errno != EAGAIN ) break;
+    const int n = write( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
+    else if( n < 0 && errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
 void xinit( pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_mutex_init( mutex, 0 );
-  if( errcode ) { show_error( "pthread_mutex_init", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_mutex_init", errcode ); cleanup_and_fail(); }
   }
 
 void xinit( pthread_cond_t * const cond )
   {
   const int errcode = pthread_cond_init( cond, 0 );
-  if( errcode ) { show_error( "pthread_cond_init", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_cond_init", errcode ); cleanup_and_fail(); }
   }
 
 
 void xdestroy( pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_mutex_destroy( mutex );
-  if( errcode ) { show_error( "pthread_mutex_destroy", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_mutex_destroy", errcode ); cleanup_and_fail(); }
   }
 
 void xdestroy( pthread_cond_t * const cond )
   {
   const int errcode = pthread_cond_destroy( cond );
-  if( errcode ) { show_error( "pthread_cond_destroy", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_cond_destroy", errcode ); cleanup_and_fail(); }
   }
 
 
 void xlock( pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_mutex_lock( mutex );
-  if( errcode ) { show_error( "pthread_mutex_lock", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_mutex_lock", errcode ); cleanup_and_fail(); }
   }
 
 
 void xunlock( pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_mutex_unlock( mutex );
-  if( errcode ) { show_error( "pthread_mutex_unlock", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_mutex_unlock", errcode ); cleanup_and_fail(); }
   }
 
 
 void xwait( pthread_cond_t * const cond, pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_cond_wait( cond, mutex );
-  if( errcode ) { show_error( "pthread_cond_wait", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_cond_wait", errcode ); cleanup_and_fail(); }
   }
 
 
 void xsignal( pthread_cond_t * const cond )
   {
   const int errcode = pthread_cond_signal( cond );
-  if( errcode ) { show_error( "pthread_cond_signal", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_cond_signal", errcode ); cleanup_and_fail(); }
   }
 
 
 void xbroadcast( pthread_cond_t * const cond )
   {
   const int errcode = pthread_cond_broadcast( cond );
-  if( errcode ) { show_error( "pthread_cond_broadcast", errcode ); fatal(); }
+  if( errcode )
+    { show_error( "pthread_cond_broadcast", errcode ); cleanup_and_fail(); }
   }
 
 
@@ -147,9 +156,11 @@ const char * const mem_msg = "Not enough memory. Try a smaller dictionary size";
 
 struct Packet			// data block with a serial number
   {
-  unsigned id;			// serial number assigned as received
   uint8_t * data;
   int size;			// number of bytes in data (if any)
+  unsigned id;			// serial number assigned as received
+  Packet( uint8_t * const d, const int s, const unsigned i )
+    : data( d ), size( s ), id( i ) {}
   };
 
 
@@ -198,10 +209,7 @@ public:
   // make a packet with data received from splitter
   void receive_packet( uint8_t * const data, const int size )
     {
-    Packet * const ipacket = new Packet;
-    ipacket->id = receive_id++;
-    ipacket->data = data;
-    ipacket->size = size;
+    Packet * const ipacket = new Packet( data, size, receive_id++ );
     slot_tally.get_slot();		// wait for a free slot
     xlock( &imutex );
     packet_queue.push( ipacket );
@@ -239,7 +247,7 @@ public:
   // collect a packet from a worker
   void collect_packet( const Packet * const opacket )
     {
-    const int i = opacket->id%num_slots;
+    const int i = opacket->id % num_slots;
     xlock( &omutex );
     // id collision shouldn't happen
     if( circular_buffer[i] != 0 )
@@ -301,6 +309,7 @@ struct Splitter_arg
   const Pretty_print * pp;
   int infd;
   int data_size;
+  int offset;
   };
 
 
@@ -313,14 +322,15 @@ extern "C" void * csplitter( void * arg )
   const Pretty_print & pp = *tmp.pp;
   const int infd = tmp.infd;
   const int data_size = tmp.data_size;
+  const int offset = tmp.offset;
 
   for( bool first_post = true; ; first_post = false )
     {
-    uint8_t * const data = new( std::nothrow ) uint8_t[data_size];
-    if( !data ) { pp( mem_msg ); fatal(); }
-    const int size = readblock( infd, data, data_size );
+    uint8_t * const data = new( std::nothrow ) uint8_t[offset+data_size];
+    if( !data ) { pp( mem_msg ); cleanup_and_fail(); }
+    const int size = readblock( infd, data + offset, data_size );
     if( size != data_size && errno )
-      { pp(); show_error( "Read error", errno ); fatal(); }
+      { pp(); show_error( "Read error", errno ); cleanup_and_fail(); }
 
     if( size > 0 || first_post )	// first packet may be empty
       {
@@ -345,6 +355,7 @@ struct Worker_arg
   const Pretty_print * pp;
   int dictionary_size;
   int match_len_limit;
+  int offset;
   };
 
 
@@ -357,15 +368,13 @@ extern "C" void * cworker( void * arg )
   const Pretty_print & pp = *tmp.pp;
   const int dictionary_size = tmp.dictionary_size;
   const int match_len_limit = tmp.match_len_limit;
+  const int offset = tmp.offset;
 
   while( true )
     {
     Packet * const packet = courier.distribute_packet();
     if( !packet ) break;		// no more packets to process
 
-    const int max_compr_size = 42 + packet->size + ( ( packet->size + 7 ) / 8 );
-    uint8_t * const new_data = new( std::nothrow ) uint8_t[max_compr_size];
-    if( !new_data ) { pp( mem_msg ); fatal(); }
     const int dict_size = std::max( LZ_min_dictionary_size(),
                                     std::min( dictionary_size, packet->size ) );
     LZ_Encoder * const encoder =
@@ -376,7 +385,7 @@ extern "C" void * cworker( void * arg )
         pp( mem_msg );
       else
         internal_error( "invalid argument to encoder" );
-      fatal();
+      cleanup_and_fail();
       }
 
     int written = 0;
@@ -387,34 +396,34 @@ extern "C" void * cworker( void * arg )
         {
         if( written < packet->size )
           {
-          const int wr = LZ_compress_write( encoder, packet->data + written,
+          const int wr = LZ_compress_write( encoder,
+                                            packet->data + offset + written,
                                             packet->size - written );
           if( wr < 0 ) internal_error( "library error (LZ_compress_write)" );
           written += wr;
           }
-        if( written >= packet->size )
-          { delete[] packet->data; LZ_compress_finish( encoder ); }
+        if( written >= packet->size ) LZ_compress_finish( encoder );
         }
-      const int rd = LZ_compress_read( encoder, new_data + new_pos,
-                                       max_compr_size - new_pos );
+      const int rd = LZ_compress_read( encoder, packet->data + new_pos,
+                                       offset + written - new_pos );
       if( rd < 0 )
         {
         pp();
         if( verbosity >= 0 )
           std::fprintf( stderr, "LZ_compress_read error: %s.\n",
                         LZ_strerror( LZ_compress_errno( encoder ) ) );
-        fatal();
+        cleanup_and_fail();
         }
       new_pos += rd;
-      if( new_pos > max_compr_size )
+      if( new_pos >= offset + written )
         internal_error( "packet size exceeded in worker" );
       if( LZ_compress_finished( encoder ) == 1 ) break;
       }
 
     if( LZ_compress_close( encoder ) < 0 )
-      { pp( "LZ_compress_close failed" ); fatal(); }
+      { pp( "LZ_compress_close failed." ); cleanup_and_fail(); }
 
-    packet->data = new_data;
+    if( packet->size > 0 ) show_progress( packet->size );
     packet->size = new_pos;
     courier.collect_packet( packet );
     }
@@ -430,19 +439,16 @@ void muxer( Packet_courier & courier, const Pretty_print & pp, const int outfd )
   while( true )
     {
     courier.deliver_packets( packet_vector );
-    if( packet_vector.size() == 0 ) break;		// all workers exited
+    if( packet_vector.empty() ) break;		// all workers exited
 
     for( unsigned i = 0; i < packet_vector.size(); ++i )
       {
       const Packet * const opacket = packet_vector[i];
       out_size += opacket->size;
 
-      if( outfd >= 0 )
-        {
-        const int wr = writeblock( outfd, opacket->data, opacket->size );
-        if( wr != opacket->size )
-          { pp(); show_error( "Write error", errno ); fatal(); }
-        }
+      const int wr = writeblock( outfd, opacket->data, opacket->size );
+      if( wr != opacket->size )
+        { pp(); show_error( "Write error", errno ); cleanup_and_fail(); }
       delete[] opacket->data;
       delete opacket;
       }
@@ -459,6 +465,7 @@ int compress( const int data_size, const int dictionary_size,
               const int infd, const int outfd,
               const Pretty_print & pp, const int debug_level )
   {
+  const int offset = data_size / 8;
   const int slots_per_worker = 2;
   const int num_slots =
     ( ( num_workers > 1 ) ? num_workers * slots_per_worker : 1 );
@@ -471,25 +478,27 @@ int compress( const int data_size, const int dictionary_size,
   splitter_arg.pp = &pp;
   splitter_arg.infd = infd;
   splitter_arg.data_size = data_size;
+  splitter_arg.offset = offset;
 
   pthread_t splitter_thread;
   int errcode = pthread_create( &splitter_thread, 0, csplitter, &splitter_arg );
   if( errcode )
-    { show_error( "Can't create splitter thread", errcode ); fatal(); }
+    { show_error( "Can't create splitter thread", errcode ); cleanup_and_fail(); }
 
   Worker_arg worker_arg;
   worker_arg.courier = &courier;
   worker_arg.pp = &pp;
   worker_arg.dictionary_size = dictionary_size;
   worker_arg.match_len_limit = match_len_limit;
+  worker_arg.offset = offset;
 
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
-  if( !worker_threads ) { pp( mem_msg ); fatal(); }
+  if( !worker_threads ) { pp( mem_msg ); cleanup_and_fail(); }
   for( int i = 0; i < num_workers; ++i )
     {
     errcode = pthread_create( worker_threads + i, 0, cworker, &worker_arg );
     if( errcode )
-      { show_error( "Can't create worker threads", errcode ); fatal(); }
+      { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
     }
 
   muxer( courier, pp, outfd );
@@ -498,13 +507,13 @@ int compress( const int data_size, const int dictionary_size,
     {
     errcode = pthread_join( worker_threads[i], 0 );
     if( errcode )
-      { show_error( "Can't join worker threads", errcode ); fatal(); }
+      { show_error( "Can't join worker threads", errcode ); cleanup_and_fail(); }
     }
   delete[] worker_threads;
 
   errcode = pthread_join( splitter_thread, 0 );
   if( errcode )
-    { show_error( "Can't join splitter thread", errcode ); fatal(); }
+    { show_error( "Can't join splitter thread", errcode ); cleanup_and_fail(); }
 
   if( verbosity >= 1 )
     {
