@@ -1,10 +1,10 @@
-/*  Plzip - A parallel compressor compatible with lzip
+/*  Plzip - Parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2009-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -43,17 +43,17 @@
 int preadblock( const int fd, uint8_t * const buf, const int size,
                 const long long pos )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = pread( fd, buf + size - rest, rest, pos + size - rest );
-    if( n > 0 ) rest -= n;
+    const int n = pread( fd, buf + sz, size - sz, pos + sz );
+    if( n > 0 ) sz += n;
     else if( n == 0 ) break;				// EOF
-    else if( errno != EINTR && errno != EAGAIN ) break;
+    else if( errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
@@ -63,16 +63,16 @@ int preadblock( const int fd, uint8_t * const buf, const int size,
 int pwriteblock( const int fd, const uint8_t * const buf, const int size,
                  const long long pos )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = pwrite( fd, buf + size - rest, rest, pos + size - rest );
-    if( n > 0 ) rest -= n;
-    else if( n < 0 && errno != EINTR && errno != EAGAIN ) break;
+    const int n = pwrite( fd, buf + sz, size - sz, pos + sz );
+    if( n > 0 ) sz += n;
+    else if( n < 0 && errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
@@ -122,9 +122,9 @@ extern "C" void * dworker( void * arg )
   LZ_Decoder * const decoder = LZ_decompress_open();
   if( !ibuffer || !obuffer || !decoder ||
       LZ_decompress_errno( decoder ) != LZ_ok )
-    { pp( "Not enough memory" ); fatal(); }
+    { pp( "Not enough memory." ); cleanup_and_fail(); }
 
-  for( int i = worker_id; i < file_index.members(); i += num_workers )
+  for( long i = worker_id; i < file_index.members(); i += num_workers )
     {
     long long data_pos = file_index.dblock( i ).pos();
     long long data_rest = file_index.dblock( i ).size();
@@ -140,11 +140,11 @@ extern "C" void * dworker( void * arg )
         if( size > 0 )
           {
           if( preadblock( infd, ibuffer, size, member_pos ) != size )
-            { pp(); show_error( "Read error", errno ); fatal(); }
+            { pp(); show_error( "Read error", errno ); cleanup_and_fail(); }
           member_pos += size;
           member_rest -= size;
           if( LZ_decompress_write( decoder, ibuffer, size ) != size )
-            internal_error( "library error (LZ_decompress_write)" );
+            internal_error( "library error (LZ_decompress_write)." );
           }
         if( member_rest <= 0 ) { LZ_decompress_finish( decoder ); break; }
         }
@@ -152,7 +152,7 @@ extern "C" void * dworker( void * arg )
         {
         const int rd = LZ_decompress_read( decoder, obuffer, buffer_size );
         if( rd < 0 )
-          fatal( decompress_read_error( decoder, pp, worker_id ) );
+          cleanup_and_fail( decompress_read_error( decoder, pp, worker_id ) );
         if( rd > 0 && outfd >= 0 )
           {
           const int wr = pwriteblock( outfd, obuffer, rd, data_pos );
@@ -160,9 +160,9 @@ extern "C" void * dworker( void * arg )
             {
             pp();
             if( verbosity >= 0 )
-              std::fprintf( stderr, "Write error in worker %d: %s\n",
+              std::fprintf( stderr, "Write error in worker %d: %s.\n",
                             worker_id, std::strerror( errno ) );
-            fatal();
+            cleanup_and_fail();
             }
           }
         if( rd > 0 )
@@ -173,7 +173,7 @@ extern "C" void * dworker( void * arg )
         if( LZ_decompress_finished( decoder ) == 1 )
           {
           if( data_rest != 0 )
-            internal_error( "final data_rest != 0" );
+            internal_error( "final data_rest is not zero." );
           LZ_decompress_reset( decoder );	// prepare for new member
           break;
           }
@@ -184,9 +184,9 @@ extern "C" void * dworker( void * arg )
 
   delete[] obuffer; delete[] ibuffer;
   if( LZ_decompress_member_position( decoder ) != 0 )
-    { pp( "Error, some data remains in decoder" ); fatal(); }
+    { pp( "Error, some data remains in decoder." ); cleanup_and_fail(); }
   if( LZ_decompress_close( decoder ) < 0 )
-    { pp( "LZ_decompress_close failed" ); fatal(); }
+    { pp( "LZ_decompress_close failed." ); cleanup_and_fail(); }
   return 0;
   }
 
@@ -196,20 +196,21 @@ extern "C" void * dworker( void * arg )
     // start the workers and wait for them to finish.
 int decompress( int num_workers, const int infd, const int outfd,
                 const Pretty_print & pp, const int debug_level,
-                const bool testing, const bool infd_isreg )
+                const bool infd_isreg )
   {
   if( !infd_isreg )
-    return dec_stream( num_workers, infd, outfd, pp, debug_level, testing );
+    return dec_stream( num_workers, infd, outfd, pp, debug_level );
 
   const File_index file_index( infd );
   if( file_index.retval() == 1 )
     {
     lseek( infd, 0, SEEK_SET );
-    return dec_stream( num_workers, infd, outfd, pp, debug_level, testing );
+    return dec_stream( num_workers, infd, outfd, pp, debug_level );
     }
   if( file_index.retval() != 0 )
-    { show_error( file_index.error().c_str() ); return file_index.retval(); }
+    { pp( file_index.error().c_str() ); return file_index.retval(); }
 
+  show_header( file_index.dictionary_size( 0 ) );
   if( num_workers > file_index.members() )
     num_workers = file_index.members();
 
@@ -224,7 +225,7 @@ int decompress( int num_workers, const int infd, const int outfd,
   Worker_arg * worker_args = new( std::nothrow ) Worker_arg[num_workers];
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
   if( !worker_args || !worker_threads )
-    { pp( "Not enough memory" ); fatal(); }
+    { pp( "Not enough memory." ); cleanup_and_fail(); }
   for( int i = 0; i < num_workers; ++i )
     {
     worker_args[i].file_index = &file_index;
@@ -236,14 +237,14 @@ int decompress( int num_workers, const int infd, const int outfd,
     const int errcode =
       pthread_create( &worker_threads[i], 0, dworker, &worker_args[i] );
     if( errcode )
-      { show_error( "Can't create worker threads", errcode ); fatal(); }
+      { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
     }
 
   for( int i = num_workers - 1; i >= 0; --i )
     {
     const int errcode = pthread_join( worker_threads[i], 0 );
     if( errcode )
-      { show_error( "Can't join worker threads", errcode ); fatal(); }
+      { show_error( "Can't join worker threads", errcode ); cleanup_and_fail(); }
     }
   delete[] worker_threads;
   delete[] worker_args;
@@ -255,11 +256,11 @@ int decompress( int num_workers, const int infd, const int outfd,
                   (double)out_size / in_size,
                   ( 8.0 * in_size ) / out_size,
                   100.0 * ( 1.0 - ( (double)in_size / out_size ) ) );
-  if( verbosity >= 3 )
+  if( verbosity >= 4 )
     std::fprintf( stderr, "decompressed size %9llu, size %9llu.  ",
                   out_size, in_size );
 
-  if( verbosity >= 1 ) std::fprintf( stderr, testing ? "ok\n" : "done\n" );
+  if( verbosity >= 1 ) std::fprintf( stderr, (outfd < 0) ? "ok\n" : "done\n" );
 
   return 0;
   }
