@@ -1,10 +1,10 @@
 /*  Plzip - Parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2009-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -59,6 +59,10 @@
 #include "arg_parser.h"
 #include "lzip.h"
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
@@ -68,15 +72,9 @@ namespace {
 
 const char * const Program_name = "Plzip";
 const char * const program_name = "plzip";
-const char * const program_year = "2013";
+const char * const program_year = "2015";
 const char * invocation_name = 0;
 
-#ifdef O_BINARY
-const int o_binary = O_BINARY;
-#else
-const int o_binary = 0;
-#endif
-
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
@@ -105,7 +103,7 @@ void show_help( const long num_online )
   std::printf( "\nOptions:\n"
                "  -h, --help                     display this help and exit\n"
                "  -V, --version                  output version information and exit\n"
-               "  -B, --data-size=<bytes>        set input data block size in bytes\n"
+               "  -B, --data-size=<bytes>        set size of input data blocks, in bytes\n"
                "  -c, --stdout                   send output to standard output\n"
                "  -d, --decompress               decompress\n"
                "  -f, --force                    overwrite existing output files\n"
@@ -121,7 +119,7 @@ void show_help( const long num_online )
                "  -1 .. -9                       set compression level [default 6]\n"
                "      --fast                     alias for -1\n"
                "      --best                     alias for -9\n", num_online );
-  if( verbosity > 0 )
+  if( verbosity >= 1 )
     {
     std::printf( "  -D, --debug=<level>        (0-1) print debug statistics to stderr\n" );
     }
@@ -132,8 +130,7 @@ void show_help( const long num_online )
                "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
                "scale optimal for all files. If your files are large, very repetitive,\n"
                "etc, you may need to use the --match-length and --dictionary-size\n"
-               "options directly to achieve optimal performance. For example, -9m64\n"
-               "usually compresses executables more (and faster) than -9.\n"
+               "options directly to achieve optimal performance.\n"
                "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
                "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
                "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
@@ -145,15 +142,37 @@ void show_help( const long num_online )
 
 void show_version()
   {
-  std::printf( "%s %s\n", Program_name, PROGVERSION );
+  std::printf( "%s %s\n", program_name, PROGVERSION );
   std::printf( "Copyright (C) 2009 Laszlo Ersek.\n"
                "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
   std::printf( "Using Lzlib %s\n", LZ_version() );
-  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  std::printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
                "This is free software: you are free to change and redistribute it.\n"
                "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
+} // end namespace
+
+void show_header( const unsigned dictionary_size )
+  {
+  if( verbosity >= 3 )
+    {
+    const char * const prefix[8] =
+      { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
+    enum { factor = 1024 };
+    const char * p = "";
+    const char * np = "  ";
+    unsigned num = dictionary_size;
+    bool exact = ( num % factor == 0 );
+
+    for( int i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
+      { num /= factor; if( num % factor != 0 ) exact = false;
+        p = prefix[i]; np = ""; }
+    std::fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
+    }
+  }
+
+namespace {
 
 unsigned long long getnum( const char * const ptr,
                            const unsigned long long llimit,
@@ -247,7 +266,7 @@ int open_instream( const char * const name, struct stat * const in_statsp,
     }
   else
     {
-    infd = open( name, O_RDONLY | o_binary );
+    infd = open( name, O_RDONLY | O_BINARY );
     if( infd < 0 )
       {
       if( verbosity >= 0 )
@@ -306,7 +325,7 @@ void set_d_outname( const std::string & name, const int i )
 
 bool open_outstream( const bool force )
   {
-  int flags = O_CREAT | O_WRONLY | o_binary;
+  int flags = O_CREAT | O_WRONLY | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename.c_str(), flags, outfd_mode );
@@ -325,7 +344,7 @@ bool open_outstream( const bool force )
 
 bool check_tty( const int infd, const Mode program_mode )
   {
-  if( program_mode == m_compress && outfd >= 0 && isatty( outfd ) )
+  if( program_mode == m_compress && isatty( outfd ) )
     {
     show_error( "I won't write compressed data to a terminal.", 0, true );
     return false;
@@ -339,6 +358,32 @@ bool check_tty( const int infd, const Mode program_mode )
   return true;
   }
 
+} // end namespace
+
+// This can be called from any thread, main thread or sub-threads alike,
+// since they all call common helper functions that call cleanup_and_fail()
+// in case of an error.
+//
+void cleanup_and_fail( const int retval )
+  {
+  // only one thread can delete and exit
+  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+
+  pthread_mutex_lock( &mutex );		// ignore errors to avoid loop
+  if( delete_output_on_interrupt )
+    {
+    delete_output_on_interrupt = false;
+    if( verbosity >= 0 )
+      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
+                    program_name, output_filename.c_str() );
+    if( outfd >= 0 ) { close( outfd ); outfd = -1; }
+    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
+      show_error( "WARNING: deletion of output file (apparently) failed." );
+    }
+  std::exit( retval );
+  }
+
+namespace {
 
      // Set permissions, owner and times.
 void close_and_set_permissions( const struct stat * const in_statsp )
@@ -346,10 +391,14 @@ void close_and_set_permissions( const struct stat * const in_statsp )
   bool warning = false;
   if( in_statsp )
     {
+    const mode_t mode = in_statsp->st_mode;
     // fchown will in many cases return with EPERM, which can be safely ignored.
-    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
-          errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
+    if( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) == 0 )
+      { if( fchmod( outfd, mode ) != 0 ) warning = true; }
+    else
+      if( errno != EPERM ||
+          fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
+        warning = true;
     }
   if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
   outfd = -1;
@@ -398,7 +447,7 @@ void Pretty_print::operator()( const char * const msg ) const
         std::fprintf( stderr, " " );
       if( !msg ) std::fflush( stderr );
       }
-    if( msg ) std::fprintf( stderr, "%s.\n", msg );
+    if( msg ) std::fprintf( stderr, "%s\n", msg );
     }
   }
 
@@ -411,7 +460,7 @@ void show_error( const char * const msg, const int errcode, const bool help )
       {
       std::fprintf( stderr, "%s: %s", program_name, msg );
       if( errcode > 0 )
-        std::fprintf( stderr, ": %s", std::strerror( errcode ) );
+        std::fprintf( stderr, ": %s.", std::strerror( errcode ) );
       std::fprintf( stderr, "\n" );
       }
     if( help )
@@ -424,69 +473,42 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void internal_error( const char * const msg )
   {
   if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
+    std::fprintf( stderr, "%s: internal error: %s\n", program_name, msg );
   std::exit( 3 );
   }
 
 
-// This can be called from any thread, main thread or sub-threads alike,
-// since they all call common helper functions that call cleanup_and_fail()
-// in case of an error.
-//
-void cleanup_and_fail( const int retval )
-  {
-  // only one thread can delete and exit
-  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
-
-  pthread_mutex_lock( &mutex );		// ignore errors to avoid loop
-  if( delete_output_on_interrupt )
-    {
-    delete_output_on_interrupt = false;
-    if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
-                    program_name, output_filename.c_str() );
-    if( outfd >= 0 ) { close( outfd ); outfd = -1; }
-    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
-      show_error( "WARNING: deletion of output file (apparently) failed." );
-    }
-  std::exit( retval );
-  }
-
-
 void show_progress( const int packet_size,
                     const Pretty_print * const p,
-                    const struct stat * const in_statsp )
+                    const unsigned long long cfile_size )
   {
-  static unsigned long long cfile_size = 0;	// file_size / 100
+  static unsigned long long csize = 0;		// file_size / 100
   static unsigned long long pos = 0;
   static const Pretty_print * pp = 0;
-  static pthread_mutex_t mutex;
+  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
 
-  if( p )					// initialize static vars
+  if( verbosity >= 2 )
     {
-    if( !pp ) xinit( &mutex );			// init mutex only once
-    pos = 0; pp = p;
-    cfile_size = ( in_statsp && S_ISREG( in_statsp->st_mode ) ) ?
-                 in_statsp->st_size / 100 : 0;
-    return;
-    }
-  if( pp )
-    {
-    xlock( &mutex );
-    pos += packet_size;
-    if( cfile_size > 0 )
-      std::fprintf( stderr, "%4llu%%", pos / cfile_size );
-    std::fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
-    pp->reset(); (*pp)();			// restore cursor position
-    xunlock( &mutex );
+    if( p )					// initialize static vars
+      { csize = cfile_size; pos = 0; pp = p; }
+    if( pp )
+      {
+      xlock( &mutex );
+      pos += packet_size;
+      if( csize > 0 )
+        std::fprintf( stderr, "%4llu%%", pos / csize );
+      std::fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
+      pp->reset(); (*pp)();			// restore cursor position
+      xunlock( &mutex );
+      }
     }
   }
 
 
 int main( const int argc, const char * const argv[] )
   {
-  // Mapping from gzip/bzip2 style 1..9 compression modes
-  // to the corresponding LZMA compression modes.
+  /* Mapping from gzip/bzip2 style 1..9 compression modes
+     to the corresponding LZMA compression modes. */
   const Lzma_options option_mapping[] =
     {
     { 1 << 20,   5 },		// -0
@@ -515,7 +537,7 @@ int main( const int argc, const char * const argv[] )
   invocation_name = argv[0];
 
   if( LZ_version()[0] != LZ_version_string[0] )
-    internal_error( "bad library version" );
+    internal_error( "bad library version." );
 
   const long num_online = std::max( 1L, sysconf( _SC_NPROCESSORS_ONLN ) );
   long max_workers = sysconf( _SC_THREAD_THREADS_MAX );
@@ -566,8 +588,7 @@ int main( const int argc, const char * const argv[] )
     const char * const arg = parser.argument( argind ).c_str();
     switch( code )
       {
-      case '0':
-      case '1': case '2': case '3': case '4':
+      case '0': case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'0']; break;
       case 'b': break;
@@ -592,7 +613,7 @@ int main( const int argc, const char * const argv[] )
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      default : internal_error( "uncaught option" );
+      default : internal_error( "uncaught option." );
       }
     } // end process options
 
@@ -607,7 +628,8 @@ int main( const int argc, const char * const argv[] )
   if( data_size <= 0 )
     data_size = 2 * std::max( 65536, encoder_options.dictionary_size );
   else if( data_size < encoder_options.dictionary_size )
-    encoder_options.dictionary_size = std::max( data_size, LZ_min_dictionary_size() );
+    encoder_options.dictionary_size =
+      std::max( data_size, LZ_min_dictionary_size() );
 
   if( num_workers <= 0 )
     num_workers = std::min( num_online, max_workers );
@@ -632,13 +654,13 @@ int main( const int argc, const char * const argv[] )
     struct stat in_stats;
     output_filename.clear();
 
-    if( !filenames[i].size() || filenames[i] == "-" )
+    if( filenames[i].empty() || filenames[i] == "-" )
       {
       input_filename.clear();
       infd = STDIN_FILENO;
       if( program_mode != m_test )
         {
-        if( to_stdout || !default_output_filename.size() )
+        if( to_stdout || default_output_filename.empty() )
           outfd = STDOUT_FILENO;
         else
           {
@@ -692,15 +714,13 @@ int main( const int argc, const char * const argv[] )
     int tmp;
     if( program_mode == m_compress )
       {
-      show_progress( 0, &pp, in_statsp );	// initialize static vars
-      if( verbosity >= 2 ) show_progress( 0 );	// show initial zero size
+      show_progress( 0, &pp, infd_isreg ? in_statsp->st_size / 100 : 0 ); // init
       tmp = compress( data_size, encoder_options.dictionary_size,
                       encoder_options.match_len_limit,
                       num_workers, infd, outfd, pp, debug_level );
       }
     else
-      tmp = decompress( num_workers, infd, outfd, pp, debug_level,
-                        program_mode == m_test, infd_isreg );
+      tmp = decompress( num_workers, infd, outfd, pp, debug_level, infd_isreg );
     if( tmp > retval ) retval = tmp;
     if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
